/**
 * Copyright (c) 2017, Arwin Technology Ltd
 * 
 * All rights reserved.
 * 
 */
/**@file
 *
 * @defgroup ble_dis_c Device Information Service Client
 * @{
 * @ingroup  ble_sdk_srv
 * @brief    Device Information Client module.
 *
 * @details  This module contains APIs to read and interact with the Device Information Service of a remote
 *           device.
 *
 * @note    The application must register this module as BLE event observer using the
 *          NRF_SDH_BLE_OBSERVER macro. Example:
 *          @code
 *              ble_dis_c_t instance;
 *              NRF_SDH_BLE_OBSERVER(anything, BLE_DIS_C_BLE_OBSERVER_PRIO,
 *                                   ble_dis_c_on_ble_evt, &instance);
 *          @endcode
 */

#ifndef BLE_DIS_C_H__
#define BLE_DIS_C_H__

#include <stdint.h>
#include "ble.h"
#include "ble_srv_common.h"
#include "ble_gattc.h"
#include "ble_db_discovery.h"
//#include "nrf_sdh_ble.h"

#ifdef __cplusplus
extern "C" {
#endif

#define BLE_DIS_C_BLE_OBSERVER_PRIO 2

/**@brief   Macro for defining a ble_dis_c instance.
 *
 * @param   _name   Name of the instance.
 * @hideinitializer
 */
#define BLE_DIS_C_DEF(_name)                                                                        \
static ble_dis_c_t _name;                                                                           \
NRF_SDH_BLE_OBSERVER(_name ## _obs,                                                                 \
                     BLE_DIS_C_BLE_OBSERVER_PRIO,                                                   \
                     ble_dis_c_on_ble_evt, &_name)


/**@brief   Macro for defining multiple ble_dis_c instances.
 *
 * @param   _name   Name of the array of instances.
 * @param   _cnt    Number of instances to define.
 */
#define BLE_DIS_C_ARRAY_DEF(_name, _cnt)                                                            \
static ble_dis_c_t _name[_cnt];                                                                     \
NRF_SDH_BLE_OBSERVERS(_name ## _obs,                                                                \
                      BLE_DIS_C_BLE_OBSERVER_PRIO,                                                  \
                      ble_dis_c_on_ble_evt, &_name, _cnt)

/**
 * @defgroup dis_c_enums Enumerations
 * @{
 */

/**@brief Device Information Service Client event type. */
typedef enum
{
    BLE_DIS_C_EVT_DISCOVERY_COMPLETE,   /**< Event indicating that the Device Information Service has been discovered at the peer. */
    BLE_DIS_C_EVT_MFS_NOTIFICATION,     /**< Event indicating that a notification of the Manufacturer String characteristic has been received from the peer. */
    BLE_DIS_C_EVT_MFS_READ_RESP,        /**< Event indicating that a read response on Manufacturer String characteristic has been received from peer. */    
    BLE_DIS_C_EVT_FWR_NOTIFICATION,     /**< Event indicating that a notification of the Firmware Revision characteristic has been received from the peer. */
    BLE_DIS_C_EVT_FWR_READ_RESP         /**< Event indicating that a read response on Firmware Revision characteristic has been received from peer. */
} ble_dis_c_evt_type_t;

/** @} */

/**
 * @defgroup dis_c_structs Structures
 * @{
 */

/**@brief Structure containing the handles related to the Device Information Service found on the peer. */
typedef struct
{
    uint16_t                mns_handle;       /**< Handle of the Manufacturer Name String characteristic as provided by the SoftDevice. */
    uint16_t                mno_handle;       /**< Handle of the Model Number String characteristic as provided by the SoftDevice. */
    uint16_t                sno_handle;       /**< Handle of the Serial Number String characteristic as provided by the SoftDevice. */
    uint16_t                hwr_handle;       /**< Handle of the Hardware Revision String characteristic as provided by the SoftDevice. */
    uint16_t                fwr_cccd_handle;  /**< Handle of the CCCD  the Firmware Revision String characteristic as provided by the SoftDevice. */    
    uint16_t                fwr_handle;       /**< Handle of the Firmware Revision String characteristic as provided by the SoftDevice. */
    uint16_t                swr_handle;       /**< Handle of the Software Revision String characteristic as provided by the SoftDevice. */
    uint16_t                sys_handle;       /**< Handle of the System ID characteristic as provided by the SoftDevice. */
    uint16_t                reg_handle;       /**< Handle of the IEEE Regulatory Cert Data List characteristic as provided by the SoftDevice. */
    uint16_t                pnp_handle;       /**< Handle of the PnP ID characteristic as provided by the SoftDevice. */
} ble_dis_c_db_t;

/**@brief Device Information Service Client Event structure. */
typedef struct
{
    ble_dis_c_evt_type_t evt_type;  /**< Event Type. */
    uint16_t conn_handle;           /**< Connection handle relevent to this event.*/
    union
    {
        ble_dis_c_db_t      dis_db;             /**< Device Information Service related handles found on the peer device. This will be filled if the evt_type is @ref BLE_DIS_C_EVT_DISCOVERY_COMPLETE.*/
        ble_srv_utf8_str_t  manufact_name_str;  /**< Manufacturer Name String. */
        ble_srv_utf8_str_t  model_num_str;      /**< Model Number String. */
        ble_srv_utf8_str_t  serial_num_str;     /**< Serial Number String. */
        ble_srv_utf8_str_t  hw_rev_str;         /**< Hardware Revision String. */
        ble_srv_utf8_str_t  fw_rev_str;         /**< Firmware Revision String. */
    } params;
} ble_dis_c_evt_t;

/** @} */

/**
 * @defgroup dis_c_types Types
 * @{
 */

// Forward declaration of the ble_dis_t type.
typedef struct ble_dis_c_s ble_dis_c_t;

/**@brief   Event handler type.
 *
 * @details This is the type of the event handler that should be provided by the application
 *          of this module in order to receive events.
 */
typedef void (* ble_dis_c_evt_handler_t) (ble_dis_c_t * p_dis_dis_c, ble_dis_c_evt_t * p_evt);

/** @} */

/**
 * @addtogroup dis_c_structs
 * @{
 */

/**@brief   Device Information Service Client structure. */
struct ble_dis_c_s
{
    uint16_t                conn_handle;     /**< Connection handle as provided by the SoftDevice. */
    ble_dis_c_db_t          peer_dis_db;     /**< Handles related to DIS on the peer*/
    ble_dis_c_evt_handler_t evt_handler;     /**< Application event handler to be called when there is an event related to the Battery service. */
};

/**@brief   Device Information Service Client initialization structure. */
typedef struct
{
    ble_dis_c_evt_handler_t evt_handler;  /**< Event handler to be called by the Device Information Service Client module whenever there is an event related to the Battery Service. */
} ble_dis_c_init_t;

/** @} */

/**
 * @defgroup dis_c_functions Functions
 * @{
 */

/**@brief      Function for initializing the Device Information Service Client module.
 *
 * @details    This function will initialize the module and set up Database Discovery to discover
 *             the Device Information Service. After calling this function, call @ref ble_db_discovery_start
 *             to start discovery once a link with a peer has been established.
 *
 * @param[out] p_ble_dis_c      Pointer to the Device Information Service client structure.
 * @param[in]  p_ble_dis_c_init Pointer to the Device Information Service initialization structure containing
 *                              the initialization information.
 *
 * @retval     NRF_SUCCESS      Operation success.
 * @retval     NRF_ERROR_NULL   A parameter is NULL.
 *                              Otherwise, an error code returned by @ref ble_db_discovery_evt_register.
 */
uint32_t ble_dis_c_init(ble_dis_c_t * p_ble_dis_c, ble_dis_c_init_t * p_ble_dis_c_init);


/**@brief     Function for handling BLE events from the SoftDevice.
 *
 * @details   This function will handle the BLE events received from the SoftDevice. If the BLE
 *            event is relevant for the Device Information Service Client module, then it is used to update
 *            interval variables and, if necessary, send events to the application.
 *
 * @note      This function must be called by the application.
 *
 * @param[in] p_ble_evt     Pointer to the BLE event.
 * @param[in] p_context     Pointer to the Battery Service client structure.
 */
//void ble_dis_c_on_ble_evt(ble_evt_t const * p_ble_evt, void * p_context);
void ble_dis_c_on_ble_evt(ble_dis_c_t * p_ble_dis_c, const ble_evt_t * p_ble_evt);


/**@brief   Function for enabling notifications on the Firmware revision characteristic.
 *
 * @details This function will enable to notification of the Battery Level characteristic at the
 *          peer by writing to the CCCD of the Battery Level Characteristic.
 *
 * @param   p_ble_dis_c Pointer to the Battery Service client structure.
 *
 * @retval  NRF_SUCCESS     If the SoftDevice has been requested to write to the CCCD of the peer.
 *          NRF_ERROR_NULL  Parameter is NULL.
 *                          Otherwise, an error code returned by the SoftDevice API @ref
 *                          sd_ble_gattc_write.
 */
uint32_t ble_dis_c_fwr_notif_enable(ble_dis_c_t * p_ble_dis_c);


/**@brief   Function for reading the Firmware revision characteristic.
 *
 * @param   p_ble_dis_c Pointer to the Battery Service client structure.
 *
 * @retval  NRF_SUCCESS If the read request was successfully queued to be sent to peer.
 */
uint32_t ble_dis_c_fwr_read(ble_dis_c_t * p_ble_dis_c);


/**@brief     Function for handling events from the database discovery module.
 *
 * @details   Call this function when getting a callback event from the DB discovery modue.
 *            This function will handle an event from the database discovery module, and determine
 *            if it relates to the discovery of Device Information service at the peer. If so, it will
 *            call the application's event handler indicating that the Device Information service has been
 *            discovered at the peer. It also populates the event with the service related
 *            information before providing it to the application.
 *
 * @param     p_ble_dis_c Pointer to the Battery Service client structure.
 * @param[in] p_evt Pointer to the event received from the database discovery module.
 *
 */
void ble_dis_on_db_disc_evt(ble_dis_c_t * p_ble_dis_c, ble_db_discovery_evt_t const * p_evt);


/**@brief     Function for assigning handles to a this instance of dis_c.
 *
 * @details   Call this function when a link has been established with a peer to
 *            associate this link to this instance of the module. This makes it
 *            possible to handle several link and associate each link to a particular
 *            instance of this module. The connection handle and attribute handles will be
 *            provided from the discovery event @ref BLE_DIS_C_EVT_DISCOVERY_COMPLETE.
 *
 * @param[in] p_ble_dis_c    Pointer to the Device Information client structure instance to associate.
 * @param[in] conn_handle    Connection handle to associated with the given Device Information Client Instance.
 * @param[in] p_peer_handles Attribute handles on the DIS server you want this DIS client to
 *                           interact with.
 */
uint32_t ble_dis_c_handles_assign(ble_dis_c_t *    p_ble_dis_c,
                                  uint16_t         conn_handle,
                                  ble_dis_c_db_t * p_peer_handles);

/** @} */ // End tag for Function group.

#ifdef __cplusplus
}
#endif

#endif // BLE_DIS_C_H__

/** @} */ // End tag for the file.
